"""
Figure out which BDF files mentioned in Main.xml is missing from NGAS.
"""
import io
import sys
import xml.etree.ElementTree as ET

import pandas
import requests
from pandas import DataFrame

import schema
from .util import groups_of
from schema import File
from schema.legacy_model import t_ngas_file_sets
from schema.ngasmodel import NGASFile
from sqlalchemy.sql import *
import argparse

def main():
    bdfs = get_bdfs()

    # 2. Join that to the NGAS BDF files
    missing_from_ngas(bdfs)

    # 3. Find the files that are in NGAS, but not in the legacy database
    missing_from_legacy(bdfs)

    # 4. Find the files that are in NGAS, but not in the new archive database
    missing_from_newarchive(bdfs)


def get_bdfs():
    parser = argparse.ArgumentParser(description='Check for BDF ingestion')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-f', metavar='FILE', dest='file', type=argparse.FileType('r'),
                       help='parses the given Main.xml file')
    group.add_argument('-p', metavar='PRODUCT', dest='product', type=str,
                       help='fetch the main table for the execblock with this product locator')
    args = parser.parse_args()

    # 1. Enumerate the BDF files in the Main.xml here
    if args.file:
        bdfs = get_bdf_list_from_file(args.file)
    else:
        bdfs = get_bdf_list_from_locator(args.product)

    return bdfs


def get_bdf_list_from_locator(locator):
    # this query gets us the Main.xml's ngas_id
    session = schema.create_session('SDM')
    query = """
    select f.ngas_id from science_products
    join filegroups f1 on science_products.filegroup_id = f1.filegroup_id
    join filegroups f2 on f1.filegroup_id = f2.parent_filegroup_id
    join files f on f2.filegroup_id = f.filegroup
    where
    science_product_locator = %s and
    filename = 'Main.xml'
    """
    ngas_id, = next(session.connection().execute(query, (locator,)))

    # now let's figure out where this is in NGAS
    query = """
    select nf.file_id,
       nf.file_version, nh.host_id || '.' || nh.domain ||  ':' || nh.srv_port as host
    from ngas_files nf
    join ngas_disks nd on nf.disk_id = nd.disk_id
    join ngas_hosts nh on nd.host_id = nh.host_id
    join (select max(file_version) as latest_version from ngas_files
          where ngas_files.file_id = %s) lv
        on lv.latest_version = nf.file_version
    where nf.file_id = %s"""
    ngas = schema.create_session('NGAS')
    file_id, file_version, host = next(ngas.connection().execute(query, (ngas_id, ngas_id)))

    # now let's run the request in stream mode and pass it to the other method
    r = requests.get('http://' + host + '/RETRIEVE', params={'file_id': file_id, 'file_version': file_version})
    bytes_expected = int(r.headers.get('Content-length'))
    if len(r.content) != bytes_expected:
        print(f'Error streaming file from NGAS: expected {bytes_expected} but obtained {len(r.content)}'
              f' (server probably needs to be rebooted)')
        sys.exit(1)
    else:
        return get_bdf_list_from_file(io.BytesIO(r.content))

def get_bdf_list_from_file(file):
    main = ET.parse(file)
    bdfs = DataFrame((i.attrib['entityId'] for i in main.getroot().iter('EntityRef')),
                     columns=['bdf_id'])
    bdfs['mangled_id'] = bdfs['bdf_id'].apply(lambda x: x.replace(':', '_').replace('/', '_') + '.bdf')
    bdfs.set_index('bdf_id')
    return bdfs


def missing_from_legacy(bdfs):
    legacy = schema.create_session('LEGACY')
    found = DataFrame([], columns=['file_id'])
    for bdf_chunk in groups_of(bdfs, 1000):
        query = select([t_ngas_file_sets.c.file_id]) \
            .where(t_ngas_file_sets.c.file_id.in_(bdf_chunk['mangled_id']))
        found = found.append(pandas.read_sql(query, legacy.connection()))
    found['found'] = True
    # 4.1 join up
    missing = bdfs.set_index('mangled_id').join(found.set_index('file_id'))
    missing = missing[missing.found.isna()]
    if missing.empty:
        print('Legacy archive is not missing any BDFs')
    for row in missing.itertuples():
        print(f'Legacy archive is missing {row.bdf_id}')


def missing_from_newarchive(bdfs):
    archive = schema.create_session('SDM')
    query = select([File.ngas_id, literal_column('true').label('found')])\
        .where(File.ngas_id.in_(bdfs['mangled_id']))
    found = pandas.read_sql(query, archive.connection())
    result = bdfs.set_index('mangled_id').join(found.set_index('ngas_id'))
    notfound = result[result.found.isna()]
    if notfound.empty:
        print('New archive is not missing any BDFs')
    for row in notfound.itertuples():
        print(f"New archive is missing {row.bdf_id}")


def missing_from_ngas(bdfs):
    ngas = schema.create_session('NGAS', profile='nmprod')
    query = select([NGASFile.file_id, literal_column('true').label('found')]) \
        .where(NGASFile.file_id.in_(bdfs['mangled_id']))
    found = pandas.read_sql(query, ngas.connection())
    # 3. Find the files that are on disk but not in NGAS
    result = bdfs.set_index('mangled_id').join(found.set_index('file_id'))
    notfound = result[result.found.isna()]
    if notfound.empty:
        print('NGAS is not missing any BDFs')
    for row in notfound.itertuples():
        print(f"BDF {row.bdf_id} is not in NGAS")
